@@ -128,6 +128,10 @@
            #:parse-cidr
            #:random-alphanumeric
            #:valid-hostname-p
+           #:prog-changes
+           #:add-change
+           #:return-changes
+           #:sh-script-to-single-line
 
            #:*consfigurator-debug-level*
            #:with-indented-inform
@@ -411,14 +415,6 @@
 
                     #:posix-capability-p))
 
-  (package :consfigurator.util.linux-namespace
-           (:use #:consfigurator.util.posix1e #:cffi)
-           (:export #:get-ids-offset
-                    #:reduce-id-maps
-                    #:shift-ids
-                    #:setgroups-p
-                    #:get-userns-owner))
-
   (package :consfigurator.property.cmd
            (:export #:single))
 
@@ -563,12 +559,23 @@
                              (#:os    #:consfigurator.property.os))
            (:export #:has-account
                     #:has-account-with-uid
+                    #:group-exists
                     #:has-groups
                     #:has-desktop-groups
 	            #:has-login-shell
                     #:has-enabled-password
                     #:has-locked-password
-	            #:passwd-field))
+	            #:passwd-field
+                    #:user-info))
+
+  (package :consfigurator.util.linux-namespace
+           (:use #:consfigurator.util.posix1e #:cffi)
+           (:local-nicknames (#:user #:consfigurator.property.user))
+           (:export #:get-ids-offset
+                    #:reduce-id-maps
+                    #:shift-ids
+                    #:setgroups-p
+                    #:get-userns-owner))
 
   (package :consfigurator.property.chroot
            (:local-nicknames (#:service   #:consfigurator.property.service)
@@ -639,6 +646,7 @@
                     #:linux-swap
 
                     #:opened-volumes
+                    #:opened-volume-parents
                     #:with-opened-volumes
 
                     #:has-volumes
@@ -736,7 +744,8 @@
                     #:chroot-installed-to-volumes-for
                     #:bootloader-binaries-installed
                     #:bootloaders-installed
-                    #:cleanly-installed-once))
+                    #:cleanly-installed-once
+                    #:with-cleanly-installed-once))
 
   (package :consfigurator.property.grub
            (:use #:consfigurator.property.disk
